import os
import pymongo
from werkzeug.security import check_password_hash

class Auth:
    def __init__(self):
        """Connect to MongoDB Atlas for authentication."""
        self.uri =  self.uri = os.getenv("MONGO_URI")
        self.client = pymongo.MongoClient(self.uri)
        self.db = self.client["erioonDB"]

    def authenticate(self, auth_string):
        """
        Authenticate using API Key, Email & Password.
        Expected format: "API_KEY/email:password"
        Returns: (authenticated: bool, user_email: str, api_key: str)
        """
        try:
            # Split auth_string into API Key and Credentials (email:password)
            api_key, credentials = auth_string.split("/")
            email, password = credentials.split(":")
        except ValueError:
            raise ValueError("Invalid authentication format. Expected 'API_KEY/email:password'.")

        # Check if the API Key is valid
        if not self.authenticate_api_key(api_key):
            print("Invalid API Key.")
            return False, None, None

        # Check if the Email exists
        if not self.authenticate_email(email):
            print("Invalid Email.")
            return False, None, None

        # Check if the Password is correct
        if not self.authenticate_password(email, password):
            print("Incorrect Password.")
            return False, None, None

        # If all checks pass, return authentication success
        print(f"Authentication successful for email: {email}")
        return True, email, api_key

    def authenticate_api_key(self, api_key):
        """Authenticate API Key by checking inside apis.{id_api}.key in MongoDB."""
        users_collection = self.db["users"]

        # Search in all users to find a matching API key
        for user in users_collection.find():
            for id_api, api_data in user.get("apis", {}).items():
                if api_data.get("key") == api_key:
                    return True
        return False

    def authenticate_email(self, email):
        """Authenticate email by checking if it exists in personal_info.email."""
        users_collection = self.db["users"]

        # Find the user by email
        user = users_collection.find_one({"personal_info.email": email})
        if user:
            return True
        print(f"Email not found: {email}")
        return False

    def authenticate_password(self, email, password):
        """Authenticate user by checking personal_info.email and personal_info.password in MongoDB."""
        users_collection = self.db["users"]

        # Find the user by email
        user = users_collection.find_one({"personal_info.email": email})

        if user:
            # Compare hashed password in the database with the provided password
            stored_hashed_password = user["personal_info"]["password"]
            if check_password_hash(stored_hashed_password, password):
                return True
            else:
                print("Incorrect password.")
                return False
        print(f"No user found for email: {email}")
        return False



import json
import avro.schema
import avro.io
import io
from azure.storage.blob import BlobServiceClient
from datetime import datetime
import os

# Azure Storage Credentials
STORAGE_ACCOUNT_NAME = "erioonstorage"  # Replace with your storage account name
STORAGE_ACCOUNT_KEY = "3OrHPLFyuwMRytBDEK+WYyl/cVkkRL9gYe0WQqMcL+S3M1RhSllKNSWTtSsI8LB/VBi5/HX/aVH4+AStiTI9Ow=="  # Replace with your storage account key
CONTAINER_NAME = "user-info"  # Blob storage container name


# Initialize Azure BlobServiceClient
blob_service_client = BlobServiceClient(account_url=f"https://{STORAGE_ACCOUNT_NAME}.blob.core.windows.net", credential=STORAGE_ACCOUNT_KEY)
container_client = blob_service_client.get_container_client(CONTAINER_NAME)

# Check if the container exists, and create it if necessary
def create_container_if_not_exists(container_client):
    try:
        # Try to get the container properties (this will raise an exception if the container doesn't exist)
        container_client.get_container_properties()
        print(f"Container '{CONTAINER_NAME}' already exists.")
    except Exception as e:
        # Container does not exist, so create it
        print(f"Container '{CONTAINER_NAME}' not found, creating it...")
        container_client.create_container()
        print(f"Container '{CONTAINER_NAME}' created.")

# Create the container if it doesn't exist
create_container_if_not_exists(container_client)

# Create a simple user schema for Avro
schema = {
    "type": "record",
    "name": "User",
    "fields": [
        {"name": "_id", "type": "string"},
        {"name": "name", "type": "string"},
        {"name": "surname", "type": "string"},
        {"name": "email", "type": "string"},
    ]
}

# Parse the schema
avro_schema = avro.schema.parse(json.dumps(schema))

# Create user data
user_data = {
    "_id": "12345",
    "name": "John Doe",
    "surname": "test",
    "email": "johndoe@example.com",
}

# Convert user data to Avro format (Binary)
def convert_to_avro(data, schema):
    bytes_writer = io.BytesIO()
    encoder = avro.io.BinaryEncoder(bytes_writer)
    writer = avro.io.DatumWriter(schema)
    writer.write(data, encoder)
    return bytes_writer.getvalue()

# Convert to Avro binary format
avro_binary_data = convert_to_avro(user_data, avro_schema)

# Define the Avro file blob name
avro_blob_name = f"user_{user_data['_id']}_avro.avro"
avro_blob_client = container_client.get_blob_client(avro_blob_name)

# Upload the Avro data to Azure Blob Storage
print(f"Uploading Avro data to Blob Storage as '{avro_blob_name}'...")
avro_blob_client.upload_blob(avro_binary_data, overwrite=True)
print(f"Avro data uploaded to {avro_blob_name}")

# Simulate saving data in WiredTiger-like format (binary file)
# Normally, WiredTiger format is managed by MongoDB, but here we'll just create a simple binary format for demonstration

wt_data = json.dumps(user_data).encode('utf-8')  # Simple binary encoding of user data (not actual WT format)

# Define the WT file blob name
wt_blob_name = f"user_{user_data['_id']}_wt.wt"
wt_blob_client = container_client.get_blob_client(wt_blob_name)

# Upload the WT data to Azure Blob Storage
print(f"Uploading WiredTiger-like data to Blob Storage as '{wt_blob_name}'...")
wt_blob_client.upload_blob(wt_data, overwrite=True)
print(f"WiredTiger-like data uploaded to {wt_blob_name}")

print("✅ User info stored in both Avro and WiredTiger-like formats in Azure Blob Storage!")


from azure.storage.blob import BlobServiceClient
import json
from ErioonDB.collection import Collection

STORAGE_ACCOUNT_NAME = "erioonappstorage"
STORAGE_ACCOUNT_KEY = "uJLDbKohah/jrHNoWVds1OnTWeHlu/9NNIzJJ9RVZqMuV+ATGfRCIXNBuxrnSWzOXxcJ0bQ98q3S+AStQoo3lg=="

class Database:
    def __init__(self, db_name, container_name):
        self.db_name = db_name
        self.container_name = container_name
        self.blob_service_client = BlobServiceClient(
            account_url=f"https://{STORAGE_ACCOUNT_NAME}.blob.core.windows.net",
            credential=STORAGE_ACCOUNT_KEY
        )
        self.container_client = self.blob_service_client.get_container_client(container_name)

        try:
            self.container_client.create_container()
        except Exception as e:
            return None

    def _get_blob_client(self, blob_name):
        """Get a blob client for the given blob name."""
        return self.container_client.get_blob_client(blob_name)

    def set_schema(self, collection_name, schema_json):
        """Save schema in Azure Blob Storage."""
        schema_blob = self._get_blob_client(f"{self.db_name}/{collection_name}/schema.json")
        try:
            schema_blob.upload_blob(json.dumps(schema_json), overwrite=True)
        except Exception as e:
            print(f"Error setting schema: {e}")

    def get_schema(self, collection_name):
        """Retrieve schema from Azure Blob Storage."""
        schema_blob = self._get_blob_client(f"{self.db_name}/{collection_name}/schema.json")
        try:
            return json.loads(schema_blob.download_blob().readall())
        except Exception as e:
            return None 

    def _save_collection(self, collection_name, data):
        """Save collection data in Azure Blob Storage."""
        collection_blob = self._get_blob_client(f"{self.db_name}/{collection_name}.avro")
        try:
            collection_blob.upload_blob(json.dumps(data), overwrite=True)
        except Exception as e:
            print(f"Error saving collection: {e}")

    def _load_collection(self, collection_name):
        """Load collection data from Azure Blob Storage."""
        collection_blob = self._get_blob_client(f"{self.db_name}/{collection_name}.avro")
        try:
            return json.loads(collection_blob.download_blob().readall())
        except Exception as e:
            return []

    def insert(self, collection_name, record):
        """Insert a record into a collection."""
        collection_data = self._load_collection(collection_name)
        collection_data.append(record)
        self._save_collection(collection_name, collection_data)

    def retrieve(self, collection_name, record_id):
        """Retrieve a record by ID."""
        collection_data = self._load_collection(collection_name)
        return next((record for record in collection_data if record.get("id") == record_id), None)

    def delete(self, collection_name, record_id):
        """Delete a record by ID."""
        collection_data = self._load_collection(collection_name)
        collection_data = [record for record in collection_data if record.get("id") != record_id]
        self._save_collection(collection_name, collection_data)

    def __getitem__(self, collection_name):
        """Access a collection by name."""
        return Collection(self, collection_name)



import threading

class ReadWriteLock:
    def __init__(self):
        self._readers = 0
        self._writer = False
        self._condition = threading.Condition()

    def acquire_read(self):
        with self._condition:
            while self._writer:
                self._condition.wait()
            self._readers += 1

    def release_read(self):
        with self._condition:
            self._readers -= 1
            if self._readers == 0:
                self._condition.notify_all()

    def acquire_write(self):
        with self._condition:
            while self._writer or self._readers > 0:
                self._condition.wait()
            self._writer = True

    def release_write(self):
        with self._condition:
            self._writer = False
            self._condition.notify_all()



import json
import avro.schema
import avro.io
import io
import random
import string
import time
from concurrent.futures import ThreadPoolExecutor
from azure.storage.blob import BlobServiceClient

# Azure Storage Credentials
STORAGE_ACCOUNT_NAME = "erioonsa"
STORAGE_ACCOUNT_KEY = "iVSFvK58XMcAWHqtSx6ETELGcLM4NTARINrp3kFPhpX50YplywBawSlqST0RYOlotnKoa8s0IpW9+AStSFvGQA=="
CONTAINER_NAME = "erioon-collection"


class ErioonClient:
    def __init__(self, api_key):
        """ Initializes session with API key and Azure Blob Storage connection """
        self.api_key = api_key
        self.blob_service = BlobServiceClient(
            account_url=f"https://{STORAGE_ACCOUNT_NAME}.blob.core.windows.net",
            credential=STORAGE_ACCOUNT_KEY
        )
        self.container_client = self.blob_service.get_container_client(CONTAINER_NAME)

        # Ensure the container exists
        try:
            self.container_client.get_container_properties()
            print(f"✅ Container '{CONTAINER_NAME}' exists.")
        except Exception as e:
            print(f"❌ Container '{CONTAINER_NAME}' does not exist, creating it now...")
            self.blob_service.create_container(CONTAINER_NAME)
            print(f"✅ Container '{CONTAINER_NAME}' created.")

    def get_database(self, db_name):
        """ Returns a database instance """
        return Database(self.blob_service, db_name)


class Database:
    def __init__(self, blob_service, db_name):
        self.db_name = db_name
        self.blob_service = blob_service

    def get_collection(self, collection_name):
        """ Returns a collection instance """
        return Collection(self.blob_service, self.db_name, collection_name)


class Collection:
    def __init__(self, blob_service, db_name, collection_name):
        self.db_name = db_name
        self.collection_name = collection_name
        self.blob_service = blob_service

        # Define Avro schema for the collection
        self.schema = avro.schema.Parse("""
        {
            "type": "record",
            "name": "SystemSettings",
            "fields": [
                {"name": "setting_name", "type": "string"},
                {"name": "value", "type": "string"}
            ]
        }
        """)
        
        # Define the blob name for the collection file
        self.blob_name = f"{self.db_name}/{self.collection_name}/collection.avro"
        self.blob_client = self.blob_service.get_blob_client(CONTAINER_NAME, self.blob_name)

    def insert_many(self, documents):
        """ Insert a batch of documents into the Azure Blob Storage in a single Avro file """
        print(f"📥 Inserting a batch of {len(documents)} documents into collection '{self.collection_name}'...")

        # Read the existing Avro data if the file already exists, else create a new file
        try:
            blob_data = self.blob_client.download_blob().readall()
            existing_data = io.BytesIO(blob_data)
            decoder = avro.io.BinaryDecoder(existing_data)
            reader = avro.io.DatumReader(self.schema)
            records = []
            while not existing_data.tell() == len(blob_data):
                records.append(reader.read(decoder))  # Read existing records
        except Exception as e:
            print("No existing file found or error reading it, creating a new one.")
            records = []

        # Append new records to existing ones
        records.extend(documents)

        # Write the updated data back to the Avro file
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer = avro.io.DatumWriter(self.schema)

        # Write all the records (existing + new) to the Avro file
        for record in records:
            writer.write(record, encoder)
        
        avro_data = bytes_writer.getvalue()

        # Save to Azure Blob Storage
        self.blob_client.upload_blob(avro_data, overwrite=True)
        print(f"✅ Batch of {len(documents)} documents stored successfully!")

    def generate_data(self, num_records):
        """ Generates fake data for testing insertion of a huge amount of data """
        data = []
        for i in range(num_records):
            setting_name = f"setting_{i}"
            value = ''.join(random.choices(string.ascii_lowercase + string.digits, k=100))  # Random 100 char string
            data.append({"setting_name": setting_name, "value": value})
        return data


# Main simulation
if __name__ == "__main__":
    print("🚀 Starting user session...")

    # Initialize the client and set up database and collection
    session = ErioonClient("123456789")
    db = session.get_database("System")
    collection = db.get_collection("system_settings")

    # Generate a large amount of data (e.g., 500,000 records)
    num_records = 1
    print(f"📊 Generating {num_records} records for insertion...")
    data = collection.generate_data(num_records)

    # Batch size for processing
    batch_size = 1
    batches = [data[i:i + batch_size] for i in range(0, len(data), batch_size)]

    # Start timer
    start_time = time.time()

    # Use ThreadPoolExecutor to insert data concurrently
    with ThreadPoolExecutor(max_workers=10) as executor:
        executor.map(collection.insert_many, batches)

    # End timer and calculate elapsed time
    end_time = time.time()
    elapsed_time = end_time - start_time
    print(f"✅ Successfully inserted {num_records} records in {elapsed_time:.2f} seconds.")



from azure.storage.blob import BlobServiceClient
from ErioonDB.auth import Auth
from ErioonDB.database import Database
import json

STORAGE_ACCOUNT_NAME = "erioonappstorage"
STORAGE_ACCOUNT_KEY = "uJLDbKohah/jrHNoWVds1OnTWeHlu/9NNIzJJ9RVZqMuV+ATGfRCIXNBuxrnSWzOXxcJ0bQ98q3S+AStQoo3lg=="

class ErioonClient:
    def __init__(self, auth_string, container_name):
        """
        Initialize ErioonClient with authentication using provided credentials.
        """
        self.auth_string = auth_string
        self.container_name = container_name
        self.blob_service_client = BlobServiceClient(
            account_url=f"https://{STORAGE_ACCOUNT_NAME}.blob.core.windows.net",
            credential=STORAGE_ACCOUNT_KEY
        )
        
        try:
            self.container_client = self.blob_service_client.get_container_client(container_name)

            if not self.container_client.exists():
                print(f"The provided user id does not exist. Please login or register at www.erioon.com/register to enable access.")
                return 
            else:
                self._authenticate(auth_string)

        except Exception as e:
            print(f"Error: {str(e)}")
            

    def _authenticate(self, auth_string):
        """Authenticate the user using the provided credentials (API_KEY/email:password)."""
        self.auth = Auth()
        self.authenticated, self.user_email, self.api_key = self.auth.authenticate(auth_string)

        if not self.authenticated:
            raise ValueError("Authentication failed. Invalid API Key or credentials.")
    
    def __call__(self, user_id):
        """Retrieve the user's database or create it if it doesn't exist."""
        if not self.container_client.exists():
            return "No ErioonDB account found. Please register at www.erioon.com/register"

        user_db_folder = f"{user_id}/" 
        user_db_blobs = self.container_client.list_blobs(name_starts_with=user_db_folder)
        
        db_found = False
        for blob in user_db_blobs:
            if blob.name.startswith(user_db_folder):
                db_found = True
                break

        if not db_found:
            print(f"Database '{user_id}' does not exist. Creating it now...")
            return self.create_database(user_id)
        else:
            return Database(user_id, self.container_name)

    def create_database(self, db_name):
        """Create a new user database folder inside the container."""
        user_db_folder = f"{db_name}/" 
        sys_settings_collection = f"{user_db_folder}sys_settings/"
        sys_settings_blob_client = self.container_client.get_blob_client(f"{sys_settings_collection}settings.json")
        
        try:
            empty_settings = {} 
            sys_settings_blob_client.upload_blob(json.dumps(empty_settings), overwrite=True)
            print(f"Database {db_name} created successfully!")
        except Exception as e:
            print(f"Error creating sys_settings collection or settings.json: {str(e)}")
        
        user_db = Database(db_name, self.container_name)
        return user_db



from setuptools import setup, find_packages

setup(
    name='ErioonDB',
    version='0.1.2.7',
    description='ErioonDB Engine',
    author='Erioon',
    author_email='hello@erioon.com',
    packages=find_packages(),  
    install_requires=[],  
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License', 
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',  
)



import json
from fastavro import writer, reader
from io import BytesIO
from azure.storage.blob import BlobClient

class Collection:
    def __init__(self, database, collection_name):
        self.database = database
        self.collection_name = collection_name
        self.container_client = self.database.container_client  # Assuming you can access this from Database
        self._ensure_collection_exists()

    def _ensure_collection_exists(self):
        """Ensure the collection folder and schema.json exists in the blob storage."""
        # First, check if the schema exists in the blob storage
        schema = self.database.get_schema(self.collection_name)

        if schema is None:
            print(f"Collection '{self.collection_name}' not found. Creating it now...")
            # If the schema doesn't exist, create an empty schema
            self._create_empty_collection()
        else:
            self.schema = schema
            self._apply_schema_to_existing_avros()

    def _create_empty_collection(self):
        """Create an empty collection folder and schema.json in Azure Blob Storage."""
        schema_blob = self.database._get_blob_client(f"{self.database.db_name}/{self.collection_name}/schema.json")

        # Create an empty schema (just an empty JSON object for now)
        empty_schema = {"type": "record", "name": "Collection", "fields": []} 
        try:
            schema_blob.upload_blob(json.dumps(empty_schema), overwrite=True)
            print(f"Collection '{self.collection_name}' created successfully.")
        except Exception as e:
            print(f"Error creating collection or schema.json: {e}")

    def set_schema(self, schema):
        """Set the schema for the collection and update the schema.json file."""
        schema_blob = self.database._get_blob_client(f"{self.database.db_name}/{self.collection_name}/schema.json")
        
        # Upload the new schema to schema.json
        try:
            schema_blob.upload_blob(json.dumps(schema), overwrite=True)
            self.schema = schema
            print(f"Schema for collection '{self.collection_name}' set successfully.")
            self._apply_schema_to_existing_avros()
        except Exception as e:
            print(f"Error setting schema: {e}")

    def _apply_schema_to_existing_avros(self):
        """Apply the schema to any existing Avro files in the collection."""
        # Fetch blobs directly from Azure Blob Storage
        blobs = self._get_collection_blobs()

        for blob in blobs:
            if blob.name.endswith('.avro'):
                print(f"Updating Avro file '{blob.name}' to match the new schema...")
                self._validate_and_apply_schema_to_avro(blob)

    def _get_collection_blobs(self):
        """Retrieve all blobs in the collection's folder."""
        blob_list = []
        try:
            blob_list = list(self.container_client.list_blobs(name_starts_with=f"{self.database.db_name}/{self.collection_name}/"))
        except Exception as e:
            print(f"Error fetching blobs: {e}")
        return blob_list

    def _validate_and_apply_schema_to_avro(self, blob):
        """Validate and apply the schema to an Avro file."""
        # Read the existing Avro file from blob storage
        avro_file = self.database._get_blob_client(blob.name)
        avro_data = avro_file.download_blob().readall()

        try:
            # Attempt to read the Avro data directly from the binary data
            avro_reader = reader(BytesIO(avro_data))  # This is correct
            records = list(avro_reader)

            # Validate if records match the schema (you can add additional validation if needed)
            writer(BytesIO(), self.schema, records)  # Apply the schema to the records
            print(f"Avro file '{blob.name}' is now valid with the updated schema.")
        except Exception as e:
            print(f"Error validating Avro file with schema: {e}")
            print(f"The file '{blob.name}' may not be a valid Avro file.")

    def _get_latest_avro_file(self):
        """Find the latest Avro file based on the naming convention."""
        blobs = self._get_collection_blobs()
        avro_files = [blob for blob in blobs if blob.name.endswith('.avro')]
        
        # Sort by file name (e.g., collection_name_1.avro, collection_name_2.avro...)
        avro_files.sort(key=lambda x: x.name)
        
        return avro_files[-1] if avro_files else None
    
    def _get_avro_record_count(self, avro_blob):
        """Return the number of records in the Avro file."""
        # Ensure avro_blob is a BlobClient, not BlobProperties
        if not isinstance(avro_blob, BlobClient):
            avro_blob = self.database._get_blob_client(avro_blob.name)

        avro_data = avro_blob.download_blob().readall()
        avro_reader = reader(BytesIO(avro_data))
        return len(list(avro_reader))


    def insert(self, record):
        """Insert a record into the collection."""
        print(f"Inserting record into collection '{self.collection_name}'...")
        
        # Find the latest Avro file and check the record count
        latest_avro_blob = self._get_latest_avro_file()
        
        # If no file exists or the latest file has more than 1000 records, create a new one
        if not latest_avro_blob or self._get_avro_record_count(latest_avro_blob) >= 1000:
            last_avro_num = len(self._get_collection_blobs())
            if last_avro_num == 1:
                new_avro_file_name = f"{self.database.db_name}/{self.collection_name}/{self.collection_name}_1.avro"
            else:
                new_avro_file_name = f"{self.database.db_name}/{self.collection_name}/{self.collection_name}_{len(self._get_collection_blobs()) + 1}.avro"
            self._create_new_avro_file(new_avro_file_name, [record])
        else:
            print("Not enterning")
            # Otherwise, just insert into the existing Avro file
            self.database.insert(self.collection_name, record)

    def _create_new_avro_file(self, file_name, records):
        """Create a new Avro file."""
        avro_file = self.database._get_blob_client(file_name)
    
        # Prepare a BytesIO buffer to hold the Avro data
        buffer = BytesIO()
    
        # Write records to the buffer
        try:
            writer(buffer, self.schema, records)
            buffer.seek(0)  # Ensure we are at the beginning of the buffer before uploading
            avro_file.upload_blob(buffer, overwrite=True)  # Upload the buffer as a blob
            print(f"Created new Avro file: {file_name}")
        except Exception as e:
            print(f"Error creating new Avro file: {e}")

    def insert_one(self, record):
        """Insert a single record into the collection."""
        print(f"Inserting a single record into collection '{self.collection_name}'...")
        try:
            self.insert(record)
            print(f"Record inserted successfully: {record}")
        except Exception as e:
            print(f"Error inserting record: {e}")

    def retrieve(self, record_id):
        """Retrieve a record from the collection by ID."""
        print(f"Retrieving record with ID {record_id} from collection '{self.collection_name}'...")
        return self.database.retrieve(self.collection_name, record_id)

    def delete(self, record_id):
        """Delete a record from the collection by ID."""
        print(f"Deleting record with ID {record_id} from collection '{self.collection_name}'...")
        self.database.delete(self.collection_name, record_id)

    def get_all(self):
        """Get all records in the collection."""
        print(f"Retrieving all records from collection '{self.collection_name}'...")
        return self.database._load_collection(self.collection_name)
